/* 
 * Copyright (C) 2008 Ondrej Stanek, ostan89<at>gmail.com 
 * http://www.ostan.cz/pocketbot
 *   
 * All rights reserved.
*/
#include <avr/io.h>
#include <avr/interrupt.h> 
#include <avr/eeprom.h> 
#include <util/delay.h> 

#include "IRsensors.h" 
#include "pocketBot_protocol.h"
#include "main.h"
#include "lineFollowing.h"



raw_sensors_t raw_sensors;

uint8_t sensors[7];

calibration_t calibration;
calibration_t EEMEM calibration_EEMEM;

uint16_t reference_voltage;

void (*callbackAfterConversion)(void);

void countRawDiff(void) {
	uint8_t n;
	for(n=0;n<7; n++) {
		if (raw_sensors.val[n]>=raw_sensors.pht[n>>1]) {
			raw_sensors.diff[n]=raw_sensors.val[n]-raw_sensors.pht[n>>1];
		} else { //val should be always bigger than pht, however, when it isn't so, we store the lowest value in diff, catching the underflow.
			raw_sensors.diff[n]=0;
		}
	}
}


void LoadCalibrationDataFromEEPROM(void) {
	eeprom_read_block(calibration, calibration_EEMEM, sizeof(calibration));
}

void WriteCalibrationDataToEEPROM(void) {
 	eeprom_write_block(calibration, calibration_EEMEM, sizeof(calibration));
}

void UpdateCalibrationOffset(void) {
	uint8_t n;
	for (n=0; n<7; n++) {
		calibration[n].offset=raw_sensors.diff[n];
	}
}

//this function requires actual offset data, therefore UpdateCalibrationOffset() should be called first
void UpdateCalibrationGain(void) {
	uint8_t n;
	for (n=0; n<7; n++) {
		calibration[n].gain=(raw_sensors.diff[n]-calibration[n].offset)/(float)255; 
	}
}

//this routine recalculate values from sensors acording to calibration data and save them to an special array
void NormalizeRawSensorData(void) {
	uint8_t n;
	for (n=0; n<7; n++) {
		int16_t val = raw_sensors.diff[n]-calibration[n].offset;
		if (val>0) {
			uint16_t temp=(float)(val)/(float)calibration[n].gain;

			if (temp>255) {
				sensors[n]=255;
			} else {
				//now we have value from interval <0,1>
				sensors[n]=temp; 
			}

		} else {
			sensors[n]=0;
		}
	}
}


void sensors_init(void) {
  
  /* IR LEDs are on when the logical state of a pin is "low". Unfortunately, pins are "low" by default.
  Therefore we need to set all the pins "high" before we configure them as outputs. 
	Moreover only one IR LED should be on at a time (to reduce current consumption)	*/
	PORTD |= 1<<PD3 | 1<<PD4 | 1<<PD5 | 1<<PD6 | 1<<PD7;	//set state of all IR LEDS to HIGH
	DDRD  |= 1<<PD3 | 1<<PD4 | 1<<PD5 | 1<<PD6 | 1<<PD7;	//set IR LED pins to be outputs

  /*
	Integrated analog to digital converter has to be turned on and configured.
	After a conversion is finished,	interrupt routine is executed to handle the result, therefore no busy delays are needed.
	In datasheet frequencies lower than 200kHz are recommended for 10bit ADC
	*/

	ADMUX = 1<<REFS0; //select AVcc as reference voltage

	ADCSRA = 1<<ADEN	//enable AD converter
				 | 1<<ADIE	//enable ADC interrupt when conversion is finished
				 | 1<<ADPS2	//setting the prescaler division factor to 64 (125kHz)
				 | 1<<ADPS1;

	callbackAfterConversion=0;
}



const struct {
	unsigned short IRled:4;				//which LED is turned on (0 for no led)
	unsigned short AnalogChannel:4; //code of next channel to measure
	uint16_t *pResult;						//pointer to memory location where the previous result will be stored
} sensorProceeding[] = {
	{PD5, 14, &reference_voltage }, //raise LED on PD5, start measuring internal reference, save previous result (it was propably the last item in table, which is internal regerence. It's better to suply the procedure with a valid pointer, becouse no pointer tests need to be processed then. However it's quite vain. )
	{PD5,  4, &reference_voltage }, // keep LED turned on, save previous result and start measuring ADC4 channel (Because we started raising up the LED diode in cycle before, the voltage is stable now.)
	{PD5,  3, &raw_sensors.val[2] },
	{PD4,  5, &raw_sensors.val[1] },
	{PD4,  3, &raw_sensors.pht[2] },
	{PD7,  4, &raw_sensors.val[0] },
	{PD7,  5, &raw_sensors.pht[1] },
	{PD6,  2, &raw_sensors.val[5] },
	{PD6,  5, &raw_sensors.pht[3] },
	{PD6,  4, &raw_sensors.val[4] },
	{PD3,  3, &raw_sensors.val[3] },
	{PD3,  2, &raw_sensors.pht[0] },
	{  0, 14, &raw_sensors.val[6] }

  /* note: when changing the channel of firs element, some changes should be also made in sensors_init(). First element should have LED OFF. */

};

#define PROC_TABLE_SIZE (sizeof(sensorProceeding)/sizeof(sensorProceeding[0]))

uint8_t proceedingStep = 0;

ISR(ADC_vect) { //routine called after an conversion is finished
  
	// switch the IR LED
 	PORTD = ( PORTD | (0xff-3) ) //turn all LEDs of (the "3" is mask of PD0, PD1 and PD2 which we don't want to affect - no LEDS are there)
					& ~(1<<sensorProceeding[proceedingStep].IRled); //switch the LED.

	// set the next analog channel to measure 
	ADMUX &= 0xf0;
	ADMUX |= 0x0f & sensorProceeding[proceedingStep].AnalogChannel;

 	//save the measured value into memory location
	*sensorProceeding[proceedingStep].pResult = ADCW; 
	
  proceedingStep++;
	if ( proceedingStep == PROC_TABLE_SIZE ) { //if we reach end of proceeding table
		proceedingStep=0; //we reset the counter; the proceeding cycle will be performed from beggining once the conversion will be (externaly) started. 
		
		sei(); //following code is not time critical		
		countRawDiff();
		NormalizeRawSensorData(); //process values acording to calibration settings
		getLinePosition(); //calculate line position
		if(callbackAfterConversion) {
			callbackAfterConversion();
			callbackAfterConversion=0;
		}
		if(pocketBotState!=FOLLOWING) {
			pbpDumpSensorData();
		} else {
			adjustMotorSpeeds();
		}
		
	} else { 
	  //if we have not reached end of table yet:
		ADCSRA |= 1<<ADSC; //start a new conversion
	}
}


