/* 
 * Copyright (C) 2008 Ondrej Stanek, ostan89<at>gmail.com 
 * http://www.ostan.cz/pocketbot
 *   
 * All rights reserved.
*/
#include <avr/io.h>
#include <avr/interrupt.h> 
#include <util/delay.h> 

#include "IRsensors.h" 

sensor_t sensors[7];

void sensors_init(void) {
  
  /* IR LEDs are on when the logical state of a pin is "low". Unfortunately, pins are "low" by default.
  Therefore we need to set all the pins "high" before we configure them as outputs. 
	Moreover only one IR LED should be on at a time (to reduce current consumption)	*/
	PORTD |= 1<<PD3 | 1<<PD4 | 1<<PD5 | 1<<PD6 | 1<<PD7;	//set state of all IR LEDS to HIGH
	DDRD  |= 1<<PD3 | 1<<PD4 | 1<<PD5 | 1<<PD6 | 1<<PD7;	//set IR LED pins to be outputs

  /*
	Integrated analog to digital converter has to be turned on and configured.
	After a conversion is finished,	interrupt routine is executed to handle the result, therefore no busy delays are needed.
	In datasheet frequencies lower than 200kHz are recommended for 10bit ADC
	*/

	ADMUX = 1<<REFS0; //select AVcc as reference voltage

	ADMUX |= 2; 	/* this should be the same channel as the first element in sensorProceeding table,
	 becouse when first conversion starts, the interrupt rutine hasn't been called yet
	 so there was no chance to set the desired channel (MUX). However this is only detail
	 and letting this out or passing a wrong value will affect only the first conversion after reset,
	 so it isn't critical at all */

	ADCSRA = 1<<ADEN	//enable AD converter
				 | 1<<ADIE	//enable ADC interrupt when conversion is finished
				 | 1<<ADSC	//start first conversion
				 | 1<<ADPS2	//setting the prescaler division factor to 64 (125kHz)
				 | 1<<ADPS1;


}



uint16_t reference_voltage;

const struct {
	unsigned short AnalogChannel:4; //code of next channel to measure
	unsigned short IRled:4;				//which LED is turned on (0 for no led)
	uint16_t *pResult;						//pointer to memory location where the previous result will be stored
} sensorProceeding[] = {
  /* note: when changing the channel of firs element, some changes should be also made in sensors_init(). First element should have LED OFF. */


	//at first we measure ambient lightning (voltage on each phototransistor without turning any IR LEDS on)
/*
	{2, 0, &sensors[6].raw_led_off }, //front sensor (ADC2)
	{3, 0, &sensors[0].raw_led_off }, //line sensors (ADC3..5)
	{4, 0, &sensors[2].raw_led_off }, //..
	{5, 0, &sensors[4].raw_led_off },*/
	/* at this point keep in mind that there are just three real phototransistors in the sensor module,
	so for interpolated sensors (1,3,5) no measuring is needed becouse it would gain us same values as for sensors 0,2,4
	But be aware of the fact that we have to copy the values manually elsewhere in code */


	//then we start measuring with LEDs ON
	/*{2, PD3, &sensors[6].raw_led_on }, //front sensor (ADC2, LED on PD3)

	{3, PD4, &sensors[0].raw_led_on }, //interpolated line module sensors
	{3, PD5, &sensors[1].raw_led_on },
	{4, PD5, &sensors[2].raw_led_on },
	{4, PD6, &sensors[3].raw_led_on },
	{5, PD6, &sensors[4].raw_led_on },
	{5, PD7, &sensors[5].raw_led_on },
*/
  /* and finaly we measure integrated 1.3V reference,
	which is later used to calculate the supply voltage Vcc (i.e. battery voltage)  */
//	{14, 0, &reference_voltage } 
											
};

#define PROC_TABLE_SIZE (sizeof(sensorProceeding)/sizeof(sensorProceeding[0]))

uint8_t proceedingStep = 0;

ISR(ADC_vect) { //routine called after an conversion is finished
  
	// switch the IR LED
	PORTD |= (0xff-3); //turn all LEDs of (the "3" is mask of PD0 and PD1 which we don't want to affect - no LEDS are there)
	PORTD &= ~(1<<sensorProceeding[proceedingStep].IRled);
	
	/*
	Be aware that no delay for LED to light up is used.
	I suppouse that the following instructions will take enaught time 
	to light the IRled fully up before the next conversion starts	
	This routine is about 106 instructions long, so the delay is about 13us (depending on compiler options)
	*/


	// set the next analog channel to measure 
	ADMUX &= 0xf0;
	ADMUX |= 0x0f & sensorProceeding[proceedingStep].AnalogChannel;

	//save the measured value into previous table element
	*sensorProceeding[
										(proceedingStep==0) ? (PROC_TABLE_SIZE-1) : (proceedingStep-1)  //when it's the first step, we will save the result to last item of table (which was processed previously)
										].pResult = ADCW; 
	
  proceedingStep++;
	if ( proceedingStep == PROC_TABLE_SIZE ) { //if we reach end of proceeding table
		proceedingStep=0; //we reset the counter; the proceeding cycle will be performed again from beggining

		/*We also need to complete the table, becouse some raw_led_off values are redudant and measuring 
		them twice would be waste of time, so they weren't comprehend in the proceeding table */
		sensors[1].raw_led_off=sensors[0].raw_led_off;
		sensors[3].raw_led_off=sensors[2].raw_led_off;
		sensors[5].raw_led_off=sensors[4].raw_led_off;
	}
	
	ADCSRA |= 1<<ADSC; //start a new conversion

}



