/* 
 * Copyright (C) 2008 Ondrej Stanek, ostan89<at>gmail.com 
 * http://www.ostan.cz/pocketbot
 *   
 * All rights reserved.
*/

/*
===================================
== NEC PROTOCOL DECODING LIBRARY ==
===================================

(Library for decoding protocol of slim remote control bought in GME.cz)


Complete protocol description with nice diagrams can be found here:
http://www.sbprojects.com/knowledge/ir/nec.htm

Library is fully asynchronous, it uses combination of external interrupt and timer output
compare interrupt, therefore external interrupt pin and one timer are required.

*/
#include <avr/io.h>
#include <avr/interrupt.h> 
#include <stdlib.h>

#include "NEC_IR_Rx.h" 
#include "remoteControl.h"

/*
Measured times for my remote:

Startbit: 68T, 8.704ms;
pause: 35T, 4.48ms;
Short: 4~5, 0.512~0.64ms
Long: 13, 1.664ms; 
*/


// timer should be capable of masuring times
// from 0.56ms to 9ms (by TCNT register only, no software enhancements)
// for FCPU=8MHz 1024 prescaler works fine, with this frequency
// timer gives 70 ticks in 9ms and 4 ticks in 0.56ms
#define TIMER_PRESCALER 1024 

// when using oscilator with lower precision (e.g. with some frequency offset),
// calibration can be set manually here:
#define CALIBRATION_FACTOR 0.99

#define TM2TCK(time)	CALIBRATION_FACTOR * time * 0.001 * (F_CPU / TIMER_PRESCALER) //convert time in ms to ticks of a timer
// times in timer ticks (computed from real times in ms) 
#define STARTBIT	TM2TCK(9)
#define PAUSE			TM2TCK(4.5)		 //pause after startbit
#define SHORT			TM2TCK(0.56)
#define LONG			TM2TCK(1.69)
#define REPEAT_PAUSE	TM2TCK(2.25)	//pause after startbit of repeat code
#define TOLERANCE	2

#ifdef RC_DEBUG
//just for debugging (to see if macros are returning correct values) 
const uint8_t Debug[] = {
	STARTBIT,
	PAUSE,
	SHORT,
	LONG,
	REPEAT_PAUSE,
	TOLERANCE
};
#endif


#define Timer2Start() TCCR2 |= 1<<CS22 | 1<<CS21 | 1<<CS20 //1024 prescaler
#define Timer2Stop()	TCCR2 &= ~(1<<CS22 | 1<<CS21 | 1<<CS20); TCNT2 = 0 //timer off

remoteCommand_t remoteCommand;

uint32_t data;
uint8_t step=0;
uint8_t repeatCode=0;

#ifdef RC_DEBUG

volatile rcError_t rcError;
volatile uint8_t errStep;
volatile uint8_t tcnt_val;

inline void CancelReceiving(rcError_t error) {
	tcnt_val=TCNT2;
	errStep=step;
	step = 0;
	repeatCode=0;
	Timer2Stop();
	rcError = error;
	cli();
}

#endif

#ifndef RC_DEBUG
	#define CancelReceiving(error)	step = 0; repeatCode=0; Timer2Stop()
#endif

/* odd code, should be removed
inline void CancelReceiving(rcError_t error) {
	step = 0; repeatCode=0; Timer2Stop();
	if (error<4) OCR1A=255;

}
*/

void remoteControl_init(void) {
  TCCR2 = 1<<WGM21;	//CTC mode. Timer will be cleared when TCNT2 meets OCR2
	TIMSK |= 1<<OCIE2;	//output compare interupt enable

	GICR |= 1<<INT0; //enable external interrupt0 (PD2)
	MCUCR |= 1<<ISC00; //any logical change on INT0 generates an interrupt request
//	PORTD |= 1<<PD2; //enable pullup //propably not needed, as the SFH5110 holds the line high when there is no signal
}

#ifdef RC_DEBUG
void Send_IRcommand(void) {
	USART_PutS("IR");
  char buffer[4];
	uint8_t n;
	for (n=0; n<4; n++) {
		itoa(remoteCommand[n],buffer,10);
		USART_Put('-');
		USART_PutS(buffer);
	}
	USART_Put(' ');
}
#endif

ISR(INT0_vect) {
	//these two commands should be processed first to gain best measuring precision
	register uint8_t time = TCNT2; //store the time of previous pause
	TCNT2 = 0; //reset the counter
	sei(); //now we can enable interrupts
	if (bit_is_set(PIND, PD2)) { // if rising edge came (processed after each burst from IR remote)
		step++; //increment the progress counter

		// no data are coded in bursts (LOW spaces),
		// the only thing we have to do here is to measure if the startbit
		// or short bit burst was long enough
		// And we only have to check minimal times, becouse exceeding times 
		// are catched by the output compare interrupt routine.

		if (step==1) { //if expecting startbit
			if (time < (STARTBIT-TOLERANCE)) { 
				CancelReceiving(SHORT_STARTBIT);
				return;
			} else {
				OCR2 = PAUSE+TOLERANCE;
			}
		} else { //if expecting short bit burst
			if (time < (SHORT-TOLERANCE))	{ 
				CancelReceiving(SHORT_BIT);
				return;
			} else {
				if (repeatCode) {
					CancelReceiving(REPEAT_CODE);
					//GIFR = 1<<INTF0; //clear external interrupt flag, so we will be able to decode future packets even during data are proccessed
					ProccesNECdata(remoteCommand, 1);
					return;
				}
				if (step==34) {
					//when receiving is completed, we store 4 bytes of data into an remoteCommand uint8_t[4] array
					//typecasting is needed to preform such operation, as it is kind of "hack", but this method spares a lot of computing time
					*(uint32_t*)&remoteCommand = data;
					CancelReceiving(RECEIVING_COMPLETED);
					//GIFR = 1<<INTF0; //clear external interrupt flag, so we will be able to decode future packets even during data are proccessed
					ProccesNECdata(remoteCommand, repeatCode);
					return;
				}
			}
			OCR2 = LONG+TOLERANCE; //a data coding space will follow
		}

	} else { //falling edge came (processed after each pause in transmission)

		// if we expected pause after startbit and it wasn't long enough
		// OR if we expected normal pause ans it wasn't long enough, then cancel the receiving process 
		// Take note that the exceeding values are caught by output compare interrupt,
		// therefore we do not have to check them here 
		
		switch (step) {
			case 0:  //if it's the first edge, we don't check anything, just start decoding
				Timer2Start();
				OCR2 = STARTBIT+TOLERANCE;
				break;
			case 1: //in second edge we are expecting pause after startbit, so let's test it
				//but in case that key is kept hold, only startbit, short pause and short burst are retransmitted
			
				if ( time < (REPEAT_PAUSE - TOLERANCE) ) {
					CancelReceiving(SHORT_PAUSE);
					return;
				}
				if ( time < (PAUSE - TOLERANCE -5) ) { //not a data message, but only repeat burst
					repeatCode=1;
				}
				OCR2 = SHORT+TOLERANCE; //a short low period should follow
				break;				
			default:
				//here it comes. We have to decode the data.
				
				//  this was my first approach, it worked fine, however there was problem becouse bytes in
				//  NEC protocol are coded with opossite endian, so I obtained reversed bytes
				//	data = data << 1; //shift the data 
				//	if (time>SHORT+TOLERANCE) data |= 1; //if it is long period, write logical 1
				
				// the solution was simple; just rotate uint32 in the oposite direction and set MSBit of MSByte to 1 instead of setting LSBit of LSByte
				data = data >> 1; //shift the data 
				if (time>SHORT+TOLERANCE) {	//if it is long period, write logical 1
					 //this sets the Most Significat Bit of the Most Significant Byte in uint32 data
					 //We are pointing at the 4th byte of uint32 directly, that is definitely more effecient
					 *(uint8_t*)((uint16_t)&data+3) |= 1<<7; 
				}

				OCR2 = SHORT+TOLERANCE; //a short low period should follow
			}
		}
}

ISR(TIMER2_COMP_vect) {
  CancelReceiving(TIME_EXCEEDED);
}
