/* 
 * Copyright (C) 2008 Ondrej Stanek, ostan89<at>gmail.com 
 * http://www.ostan.cz/pocketbot
 *   
 * All rights reserved.
*/
//#define DEBUG_PBP

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include "uart.h"
#include "IRsensors.h"
#include "lineFollowing.h"

#define PBP_DELIMITER 0x7e //starting delimiter of an packet

#define BUFF_SIZE_MASK 0b00111111 //64 bytes buffer

uint8_t buffer[BUFF_SIZE_MASK+1];
uint8_t btmPtr=0;
uint8_t topPtr=0; //relative pointers to top and bottom of the buffer

typedef void packetHandler_t(const uint8_t data[]);

//nonblocking function forms and sends pocketBot protocol packet
void pbpSendPacket(uint8_t command, uint8_t len, uint8_t data[]) {
	TxBufAppend(PBP_DELIMITER);
	TxBufAppend(command);

	uint8_t n;
	uint8_t checksum = command;
	for (n=0; n<len; n++) {
		TxBufAppend(data[n]);		
		checksum+=data[n];
	}
	checksum = ~(checksum+1); //inc and invert
	TxBufAppend(checksum);

}

void pbAskForCalibrationData(const uint8_t data[]) {
	pbpSendPacket(0x15,sizeof(calibration),(void*)&calibration);
}

void pbAskForRegulatorData(const uint8_t data[]) {
	pbpSendPacket(0x22,sizeof(regulator),(void*)&regulator);
}

void pbAskForFreshData(const uint8_t data[]) {
	startConversion();
}

void pbpDumpSensorData(void) {
	pbpSendPacket(0x18,sizeof(sensors),(void*)&sensors);
	pbpSendPacket(0x21,sizeof(line),(void*)&line);
	pbpSendPacket(0x19,sizeof(raw_sensors),(void*)&raw_sensors);
	pbpSendPacket(0x20,sizeof(reference_voltage),(void*)&reference_voltage);
}


typedef struct {
	uint8_t command;
	uint8_t len;
	packetHandler_t *handler;
} packetType_t;

//Table containing characteristics of packets processed by this device 
//(it isn't necessary to describe all packets of the pocketBot protocol here,
//only the packets significant for this device are published)
const packetType_t packetTable[] = 
{
	{ 0x14, 0, pbAskForCalibrationData },
	{ 0x16, 0, pbAskForFreshData },
	{ 0x30, 0, pbAskForRegulatorData }
};

#define INC(ptr) \
		(ptr)++; \
		ptr&=BUFF_SIZE_MASK	//corp with the buffer size (circular buffer made easily :)


#define ADD(ptr,count)	(((ptr)+(count))&BUFF_SIZE_MASK)

void PBP_bufferByte(uint8_t byte) {
	buffer[topPtr]=byte;
	INC(topPtr);
	if (btmPtr==topPtr) { //if the buffer owerflows, overwrite the oldest byte
		#ifdef DEBUG_PBP
		TxBufAppend(0x01);//would be nice to put here some warning that data flow is too fast and the host doesn't manage to procces the packets
		#endif
		INC(btmPtr);
	}
} 

inline uint8_t bytesInBuffer(void) {
	return ((topPtr-btmPtr)&BUFF_SIZE_MASK);
}

const packetType_t *getPacketType(uint8_t id) {
	//search for a packet with certain id, return 0 if such id isn't in table
	uint8_t n;
	for(n=0; n<(sizeof(packetTable)/sizeof(packetType_t)); n++) {
		if (packetTable[n].command==id) {
			return &(packetTable[n]);
		}
	}
	return 0;
}

//counts checksum and if it's valid, call packet handler function
//return 0 if checksum
void processPacket(const uint8_t data[], const packetType_t *type) {
	#ifdef DEBUG_PBP
	TxBufAppend(0x02);
	#endif
	uint8_t n;
	uint8_t checksum = type->command; //command is included in checksum
	for (n=0; n<(type->len); n++) {
		checksum+=data[n];
	}
	checksum = ~(checksum+1); //inc and invert

	#ifdef DEBUG_PBP
	TxBufAppend(10);
	TxBufAppend(checksum);
	TxBufAppend(data[type->len]);
	#endif

	if (checksum == data[type->len]) {
		(type->handler)(data); //if packet was OK, call handler function
	} else {
		#ifdef DEBUG_PBP
		TxBufAppend(0x03);//some error code could be send to controller, not implemented yet
		#endif
	}

}

//this non-blocking routine reassembles packets form data buffer
//when packet is sucesfully decoded, callback to packet handler function is made immediately
void pocketBotBufferPoll(void) { 
	while ( (bytesInBuffer()>0) && (buffer[btmPtr]!=PBP_DELIMITER) ) {
		INC(btmPtr);
		#ifdef DEBUG_PBP
		TxBufAppend(0x05);
		#endif
	} //at this point we have either the beginning of packet or the buffer is empty
	if (bytesInBuffer()>=3) { //if there is enaugh bytes to form a potentional packet (the shortest packet has three bytes: DELIMITER, COMMAND and CHECKSUM)
			#ifdef DEBUG_PBP
			TxBufAppend(0x08);
			#endif
	
		const packetType_t *type = getPacketType(buffer[ADD(btmPtr,1)]);	//lets look if the command is in our table		
		if (type!=0) { //if command was understood
			#ifdef DEBUG_PBP
			TxBufAppend(0x06);
			#endif
			if ((type->len+3) <= bytesInBuffer()) { //and if the whole packet arrived (adding 3 for DELIMITER, COMMAND and CHECKSUM, becouse len counts only data bytes)
				if (ADD(btmPtr,2)<=ADD(btmPtr,2 + (type->len) + 1)) { //if data are consistent, we can pass pointer to array directly (including the checksum at the end of packet)
					processPacket(&buffer[ADD(btmPtr,2)], type);
					btmPtr=ADD(btmPtr,type->len+3); //packet was processed, so we need to move the pointer on next packet
				}	else { //if data aren't consistent, we have to copy them to a buffer
					uint8_t *argBuff;
					argBuff=(uint8_t*)malloc((type->len)+1); //allocate memory for packet (message length plus one byte for checksum)
					if (argBuff==0) { //if an error during allocation ocured, program should not continue!
						#ifdef DEBUG_PBP
						TxBufAppend(0x04);//some error code could be send to controller, not implemented yet
						#endif
					} 
					uint8_t n;
					btmPtr=ADD(btmPtr,2); //move pointer to first data byte
					for (n=0; n<(type->len+1); n++) { //fill the buffer with data plus the last byte of checksum
						argBuff[n]=buffer[btmPtr];	
						INC(btmPtr); //pointer is increased as data are being processed
					}
					processPacket(argBuff, type);
					free(argBuff); //and finally we can free the memory, becouse data are no longer needed
				}			
			}
		} else {	//when misunderstood, drop the fake delimiter
			#ifdef DEBUG_PBP
			TxBufAppend(0x07);
			#endif

			INC(btmPtr); 
			pocketBotBufferPoll(); //and try the sequence from beginning
		}
	}
}




